#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <afblib/mpx_session.h>
#include <afblib/strhash.h>
#include <afblib/strhash.h>
#include <mxp.h>
#include <mxprequest.h>
#include <mxpresponse.h>

/* users that have not registered themselves with a login name,
   have a status of START;
   after the registration the status turns to REGISTERED;
   whenever a client is blocked as it waits for a semaphore
   to become free, we switch temporarily to WAITING;
   as long as a client waits, all incoming requests will
   be responded to with error messages */
enum mpxstatus {START, REGISTERED, WAITING};

/* per-session user structure */
struct user {
   session* s; /* back link to the associated session */
   char* login; /* is 0 if status == START */
   enum mpxstatus status;
   strhash locked;
   struct semaphore* awaited;
};

/* linear list of users that wait for a semaphore to become free */
struct semaqueue {
   struct user* user;
   struct semaqueue* next;
};

/* semaphore objects exist only when they are held,
   whenever a semaphore becomes free, it gets disposed */
struct semaphore {
   char* name; /* name of this semaphore */
   struct user* holder; /* current lock holder */
   /* queue */
   struct semaqueue* head;
   struct semaqueue* tail;
};

strhash users; /* hash of (login name, struct user*) pairs */
strhash semas; /* hash of (semaphore name, struct semaphore*) pairs */

#define EQUAL(sa,str) (strncmp((sa.s),(str),(sa.len)) == 0)

/* release a semaphore:
   if the queue is non-empty: assign the semaphore to the next in the queue;
   otherwise clear all references to this semaphore and dispose it */
void release(struct semaphore* sema) {
   /* remove it from the list of semaphores held by this user */
   strhash_remove(&sema->holder->locked, sema->name);
   if (sema->head) {
      /* remove first element of the queue */
      struct semaqueue* old = sema->head;
      sema->head = old->next;
      if (sema->head == 0) {
	 sema->tail = 0;
      }
      sema->holder = old->user;
      free(old);
      /* notify new lock holder */
      mxp_response response = {MXP_SUCCESS};
      stralloc_copys(&response.message, "locked");
      if (!write_mxp_response(sema->holder->s, &response)) {
	 close_session(sema->holder->s);
      }
      send_response(sema->holder->s);
      /* update bookkeeping for the new lock holder */
      sema->holder->awaited = 0;
      int ok = strhash_add(&sema->holder->locked, sema->name,
	 (char*) sema);
      assert(ok);
      sema->holder->status = REGISTERED;
   } else {
      /* remove semaphores that became free */
      strhash_remove(&semas, sema->name);
      free(sema->name);
      free(sema);
   }
}

/* handle incoming requests */
void handle_request(session* s) {
   /* handle new sessions */
   if (s->handle == 0) {
      struct user* user =
	 (struct user*) malloc(sizeof(struct user));
      if (!user) return;
      user->s = s;
      user->login = 0;
      user->status = START;
      strhash_alloc(&user->locked, 10);
      user->awaited = 0;
      s->handle = (void*) user;
   }
   struct user* user = (struct user*) s->handle;

   /* parse request */
   mxp_request request = {{0}};
   if (!read_mxp_request(s, &request)) {
      close_session(s); return;
   }
   /* append a null-byte as we need it */
   stralloc_0(&request.parameter);

   mxp_response response = {MXP_SUCCESS}; /* be optimistic... */
   struct semaphore* sema;

   switch (user->status) {
      case START:
	 if (EQUAL(request.keyword, "id")) {
	    char* login = request.parameter.s;
	    if (strhash_add(&users, login, (char*) user)) {
	       user->login = strdup(login);
	       stralloc_copys(&response.message, "welcome");
	       user->status = REGISTERED;
	    } else {
	       response.status = MXP_FAILURE;
	       stralloc_copys(&response.message, "name already in use");
	    }
	 } else {
	    response.status = MXP_FAILURE;
	    stralloc_copys(&response.message, "please login first");
	 }
	 break;
      case REGISTERED:
	 /* fetch semaphore structure that is associated with the parameter */
	 if (!strhash_lookup(&semas, request.parameter.s, (char**) &sema)) {
	    sema = 0;
	 }
	 if (EQUAL(request.keyword, "stat")) {
	    if (sema) {
	       /* send continuation line with the name of the holder */
	       response.status = MXP_CONTINUATION;
	       stralloc_copys(&response.message, sema->holder->login);
	       if (!write_mxp_response(s, &response)) {
		  close_session(s); return;
	       }
	       send_response(s);
	       /* prepare final response line */
	       response.status = MXP_SUCCESS;
	       stralloc_copys(&response.message, "held");
	    } else {
	       stralloc_copys(&response.message, "free");
	    }
	 } else if (EQUAL(request.keyword, "lock")) {
	    if (sema) {
	       /* put ourself into the semaqueue */
	       struct semaqueue* member =
		  (struct semaqueue*) malloc(sizeof(struct semaqueue));
	       if (!member) {
		  close_session(s); return;
	       }
	       member->user = user; member->next = 0;
	       if (sema->tail) {
		  sema->tail->next = member;
	       } else {
		  sema->head = member;
	       }
	       sema->tail = member;
	       user->awaited = sema;
	       /* notify client */
	       response.status = MXP_CONTINUATION;
	       stralloc_copys(&response.message, "waiting");
	       user->status = WAITING;
	    } else {
	       /* create new semaphore */
	       sema = (struct semaphore*) malloc(sizeof(struct semaphore));
	       if (!sema) {
		  close_session(s); return;
	       }
	       sema->holder = user;
	       sema->head = 0; sema->tail = 0;
	       sema->name = strdup(request.parameter.s);
	       int ok = strhash_add(&semas, sema->name, (char*) sema);
	       assert(ok);
	       ok = strhash_add(&user->locked, sema->name, (char*) sema);
	       assert(ok);
	       /* notify client */
	       stralloc_copys(&response.message, "locked");
	    }
	 } else if (EQUAL(request.keyword, "release")) {
	    if (!sema || sema->holder != user) {
	       response.status = MXP_FAILURE;
	       stralloc_copys(&response.message, "not held by you");
	    } else {
	       release(sema);
	       stralloc_copys(&response.message, "");
	    }
	 } else {
	    response.status = MXP_FAILURE;
	    stralloc_copys(&response.message, "unknown command");
	 }
	 break;
      case WAITING:
	 response.status = MXP_FAILURE;
	 stralloc_copys(&response.message, "please continue to wait");
   }
   /* store response in our session handle s */
   if (!write_mxp_response(s, &response)) {
      close_session(s); return;
   }
}

/* handle closed connections */
void handle_close(session* s) {
   if (s->handle) {
      struct user* user = (struct user*) s->handle;
      /* release all locked semaphores */
      char* name;
      for(;;) {
	 /* start always from the beginning as this
	    iterator is not robust regarding removals */
	 strhash_start(&user->locked);
	 if (!strhash_next(&user->locked, &name)) break;
	 struct semaphore* sema;
	 int ok = strhash_lookup(&user->locked, name, (char**) &sema);
	 assert(ok);
	 release(sema);
      }
      /* remove ourself out of a waiting queue, if necessary */
      if (user->awaited) {
	 struct semaphore* sema = user->awaited;
	 struct semaqueue* prev = 0;
	 struct semaqueue* q = sema->head;
	 while (q && q->user != user) {
	    prev = q;
	    q = q->next;
	 }
	 assert(q);
	 if (prev) {
	    prev->next = q->next;
	 } else {
	    sema->head = q->next;
	 }
	 if (sema->tail == q) {
	    sema->tail = prev;
	 }
      }
      strhash_free(&user->locked);
      /* remove user from the hash table of users */
      if (user->login) {
	 strhash_remove(&users, user->login);
      }
      /* dispose user structure */
      free(user->login);
      free(user);
   }
}

int main() {
   strhash_alloc(&users, 20);
   strhash_alloc(&semas, 20);
   run_mpx_service(MXP_PORT, handle_request, handle_close);
}
